import pandas as pd
import numpy as np
from datetime import datetime
import time

class DataModel:
    def __init__(self):
        self.data = None
        self.customerId = 'customerId'
        self.orderId = 'orderId'
        self.orderDate = 'orderDate'
        self.salesPrice = 'salesPrice'

    @property
    def clear_cach(self):
        self.data = None

    @property
    def echo(self):
        print('DataModel')

    def get_range(self, x):
        return x.max()-x.min()

    def get_q1(self, x):
        return x.quantile(.25)

    def get_q3(self, x):
        return x.quantile(.75)

    def convert_dtypes(self):
        convert_dict = {'object':['customerId','orderId'],
                        'float':'salesPrice'}
        for datatype, col in convert_dict.items():
            self.data[col] = self.data[col].astype(datatype)

    def make_date_normal(self):
        self.data['date'] = self.data[self.orderDate].dt.date
        self.data['year'] = self.data[self.orderDate].dt.year
        self.data['quarter'] = self.data[self.orderDate].dt.quarter
        self.data['month'] = self.data[self.orderDate].dt.month
        self.data['week'] = self.data[self.orderDate].dt.isocalendar().week
        self.data['day'] = self.data[self.orderDate].dt.day
        self.data['dayOfYear'] = self.data[self.orderDate].dt.dayofyear

    def make_date_combined(self):
        self.data['yearQuarter'] = self.data['year'].astype(str)+'-'+self.data['quarter'].astype(str)
        self.data['yearMonth'] = self.data['year'].astype(str)+'-'+self.data['month'].astype(str).apply(lambda x: x.zfill(2))
        self.data['yearWeek'] = self.data['year'].astype(str)+'-'+self.data['week'].astype(str).apply(lambda x: x.zfill(2))
        self.data['yearDay'] = self.data['year'].astype(str)+'-'+self.data['day'].astype(str).apply(lambda x: x.zfill(2))
        self.data['yearDayYear'] = self.data['year'].astype(str)+'-'+self.data['dayOfYear'].astype(str).apply(lambda x: x.zfill(3))

    def get_date(self, date):

        combined_col = ['customerId', 'orderId', 'salesPrice', 'orderDate', 'yearQuarter',
       'yearMonth', 'yearWeek', 'yearDay', 'yearDayYear']

        if date == 'normal':
            self.make_date_normal()
        elif date == 'combined':
            self.make_date_combined()
            self.data[combined_col] = self.data[combined_col]
        elif date == 'full':
            self.make_date_normal()
            self.make_date_combined()
        else:
            pass

    def aggregate(self, data, customerId, orderId, orderDate, salesPrice):
        self.data = data.groupby([customerId,orderId]).agg(
            salesPrice = pd.NamedAgg(column=salesPrice, aggfunc='sum'),
            orderDate = pd.NamedAgg(column=orderDate, aggfunc=lambda x: x.iloc[-1]),
        ).reset_index()
        self.data.rename(columns={customerId:self.customerId,orderId:self.orderId},inplace=True)

    def prep(self, data, customerId, orderId, orderDate, salesPrice, date='full', method=None):
        """
        method = yqmwd, yq, ym, yw, yd
        """
        start = time.time()
        self.aggregate(data, customerId, orderId, orderDate, salesPrice)
        self.get_date(date)
        self.convert_dtypes()
        final_df = self.data.copy()
        self.clear_cach
        
        for col, map in zip([customerId, orderId, orderDate, salesPrice],
        [self.customerId, self.orderId, self.orderDate, self.salesPrice]):
            print(f"Changed from {col} to {map}")
        
        print(f'DataModel successfully executed {time.time()-start:.2f} ms')

        return final_df


import pandas as pd
import numpy as np
from datetime import datetime
import time
# from .DataModel import DataModel
from .EzCohort import EzCohort

class EzCustomerMovement(EzCohort):
    def __init__(self):
        super().__init__()
        self.data = 'EzCustomerMovement'

    @property
    def clear_cach(self):
        self.data = None

    @property
    def echo(self):
        print('EzCustomerMovement')

    def status_code(self, x):
        if x == 0:
            return 'new'
        elif x == 1:
            return 'repeat'
        elif x > 1:
            return 'reactivate'

    def get_status(self):
        self.data['shift'] = self.data.groupby('customerId')[self.cohortIdx].shift(1).fillna(1).astype('int')
        self.data['statusId'] = self.data[self.cohortIdx] - self.data['shift']
        self.data['status'] = self.data['statusId'].apply(self.status_code)

    def get_churn(self):
        self.data = pd.pivot_table(self.data, 
                                        index=self.basketDate,columns='status',values='customerId',
                                        aggfunc='count',fill_value=0)
        self.data['not_new'] = self.data['reactivate'] + self.data['repeat']
        self.data['cumsum'] = self.data['new'].cumsum().fillna(0)
        self.data['churn'] = self.data['new'] + self.data['not_new']-self.data['cumsum']
        self.data = self.data.reset_index()    

    def prep_customer_movement(self, data):
        self.data = self.prep_cohort(data)
        self.get_status()
        self.get_churn()
        return self.data

    def get_melt(self):
        self.data = pd.melt(self.data, id_vars=self.basketDate)
        return self.data

    def fit(self, data, method=None, melt=True):
        """
        method = normal, melt
        """
        start = time.time()
        columns = [self.basketDate,'new','reactivate','repeat','churn']
        self.data = self.prep_customer_movement(data)[columns]
        
        if melt==True:
            self.data = self.get_melt()

        final_df = self.data
        
        self.clear_cach
        print(f'EzCustomerMovement successfully executed {time.time()-start:.2f} ms')
        return final_df


import pandas as pd
import numpy as np
from datetime import datetime
import time
# from .DataModel import DataModel

class EzCohort:
    def __init__(self):
        self.data = 'EzCohort'
        self.basketDate = 'basketDate'
        self.customerId = 'customerId'
        self.orderId = 'orderId'
        self.orderDate = 'orderDate'
        self.salesPrice = 'salesPrice'
        self.cohortGrp = 'cohortGrp'
        self.cohortIdx = 'cohortIdx'
        self.frequency = 'frequency'

    @property
    def clear_cach(self):
        self.data = None

    @property
    def echo(self):
        print('EzCohort')

    def get_month(self, x):
        """
        return: datetime object
        """
        return datetime(x.year, x.month, 1)

    def get_date_int(self, column: str):
        """
        column <- column name containing invoice or order date
        return: tuple
        """
        series = pd.to_datetime(self.data[column])
        year = series.dt.year
        month = series.dt.month
        day = series.dt.day
        return year, month, day

    def get_cohort_index(self):
        basket_y, basket_m, basket_d = self.get_date_int(self.basketDate)
        cohort_y, cohort_m, cohort_d = self.get_date_int(self.cohortGrp)
        cohort_idx = (12*(basket_y - cohort_y) + (basket_m - cohort_m)) + 1
        self.data[self.cohortIdx] = cohort_idx

    def prep_cohort(self, data, method=None):
        self.data = data.copy()
        self.data[self.basketDate] = self.data[self.orderDate].apply(self.get_month)
        self.data = self.data.groupby([self.customerId, self.basketDate]).agg(
            frequency=pd.NamedAgg(column=self.orderId, aggfunc='nunique'),
            salesPrice=pd.NamedAgg(column=self.salesPrice, aggfunc='sum')
        ).reset_index()
        self.data[self.cohortGrp] = self.data.groupby(self.customerId)[self.basketDate].transform('min')
        self.get_cohort_index()
        self.data[self.customerId] = self.data[self.customerId].astype('object')
        return self.data

    def get_cohort_count(self):
        self.data = pd.pivot_table(self.data, index=self.cohortGrp,columns=self.cohortIdx, values=self.customerId, aggfunc='nunique')
        return self.data

    def get_cohort_retention(self):
        self.data = self.data.divide(self.data.iloc[:,0], axis=0)
        return self.data

    def get_cohort_sales(self):
        res = pd.pivot_table(self.data, index=self.cohortGrp,columns=self.cohortIdx, values=self.salesPrice, aggfunc='sum')
        return res

    def get_cohort_frequency(self):
        res = pd.pivot_table(self.data, index=self.cohortGrp,columns=self.cohortIdx, values=self.frequency, aggfunc='sum')
        return res

    def get_cohort_mean(self):
        total = self.get_cohort_sales()
        freq = self.get_cohort_frequency()
        self.data = total.divide(freq)
        return self.data

    def get_melt(self):
        self.data = self.data.reset_index()
        self.data = pd.melt(self.data, id_vars=self.cohortGrp)
        return self.data

    def fit(self, data, method='count', melt=False):
        """
        method = retention_rate, qty, sales, avg
        """
        start = time.time()
        self.prep_cohort(data, method=None)
        if method=='count':
            self.data = self.get_cohort_count()
        elif method=='retention':
            self.data = self.get_cohort_count()
            self.data = self.get_cohort_retention()
        elif method=='sales':
            self.data = self.get_cohort_sales()
        elif method=='frequency':
            self.data = self.get_cohort_frequency()
        elif method=='mean':
            self.get_cohort_mean()
        else:
            self.data = self.data

        if melt==True:
            self.get_melt()

        final_df = self.data
        
        self.clear_cach
        print(f'EzCohort successfully executed {time.time()-start:.2f} ms')

        return final_df.round(2)